
#include <chrono>
#include <iostream>
#include <fstream>

#include <opencv2/opencv.hpp>
#include <opencv2/videoio/videoio.hpp>

#include "dfxcpp/DFXFactory.h"
#include "dfx/Collector.h"

#include "nlohmann/json.h"
using json = nlohmann::json;

using dfxcpp::DFXFactory;
using dfx::Face;
using dfx::VideoFrame;

std::map<std::string, dfx::Face> loadFacePoints(const std::string& jsonDataFile);
dfx::Face decodeJsonFace(json& jsonFace);
void sendNetworkChunk(std::shared_ptr<dfx::Collector> collector, std::string name);

cv::Mat annotateImage(cv::Mat& image, dfx::Face& face);

#ifdef _WIN32
#define PATH_SEP "\\"
#else
#define PATH_SEP "/"
#endif

int main(int argc, char **argv) {
    std::string videoTestFile("res" PATH_SEP "dfxDemoCP1.mov");
    std::string jsonFaceDataFile("res" PATH_SEP "dfxDemoCP1-faces.json");

    std::ifstream testVideoFile(videoTestFile);
    if ( !testVideoFile ) {
        std::cerr << "Unable to locate test video file: " << videoTestFile << std::endl;
        return -1;
    }
    std::ifstream testJsonDataFile(jsonFaceDataFile);
    if ( !testJsonDataFile ) {
        std::cerr << "Unable to locate json test data file: " << jsonFaceDataFile << std::endl;
        return -1;
    }

    std::cout << "Load the face point json file" << std::endl;

    // load the pre-found faces, usually this would be done real time to
    // construct a dfx::Face for each frame but here we use pre-caculated
    // points.
    auto dfxFaces = loadFacePoints(jsonFaceDataFile);


    std::cout << "Starting video processing..." << std::endl;

    cv::VideoCapture videoFile(videoTestFile);

    DFXFactory factory;

    // Initialize where the collector configuration and models are
    // It will load the configuration file, res/deepaffex.lua to
    // obtain references to the models from res/models/*.
    if ( !factory.initializeStudyFromFile( "res" PATH_SEP "deepaffex.lua" ) ) {
        std::cerr << "DFXFactory ERROR: " << factory.getLastErrorMessage() << std::endl;
        return -1;
    }

    // This will be a continuous collector, by default it is "discrete"
    factory.setMode("continuous");

    // Now that the type of collector is setup, create a collector matching the configuration and type
    auto collector = factory.createCollector();
    if ( collector->getCollectorState() == dfx::CollectorState::Error ) {
        std::cerr << "DFXCollector creation ERROR: " << collector->getLastErrorMessage() << std::endl;
        return -1;
    }

    std::vector<uint8_t> createMeasurementResponsePayload;
    collector->prepareMeasurement(createMeasurementResponsePayload);

    auto startTime = std::chrono::high_resolution_clock::now();
    auto now = std::chrono::high_resolution_clock::now();

    unsigned long frameNumber = 1;
    auto totalFrames = videoFile.get(cv::CAP_PROP_FRAME_COUNT );
    auto fps = videoFile.get(cv::CAP_PROP_FPS);
    auto width = videoFile.get(cv::CAP_PROP_FRAME_WIDTH);
    auto height = videoFile.get(cv::CAP_PROP_FRAME_HEIGHT);

    collector->setTargetFPS(fps);       // What fps was the video encoded?
    collector->setChunkDurationSeconds(5.0f);

    std::cout << "Number Frames: " << totalFrames << "@" << fps << "fps [" << width << "x" << height << "]" << std::endl;

    // Start the collection cycle
    collector->startCollection();

    for(;;) {
        cv::Mat image;
        videoFile >> image;
        if (image.empty())
            break;

        auto timestamp_ns = static_cast<unsigned long long>(frameNumber * 1e9 / fps);

        if ((frameNumber % 100) == 0 ) {
           std::cout << "  Frame[" << frameNumber << "] @" << timestamp_ns << std::endl;
        }
        dfx::VideoFrame videoFrame {
            image,
            dfx::VideoFrame::ChannelOrder::BGR,
            timestamp_ns,
            frameNumber++
        };

        // Let the collector know about the new video frame
        auto frame = collector->createFrame(videoFrame);

        // Obtain the face points
        dfx::Face face = dfxFaces[std::to_string(frameNumber)];

        // Add the faces to the frame
        auto dfxFace = std::make_shared<dfx::Face>(face);
        frame->addFace(dfxFace);

        // Calculate the regions of interest
        collector->defineRegions(frame);

        // Filter and extract the region information
        auto result = collector->extractChannels(frame);

        // The video may be chunked if it is large, if the chunk
        // is ready we need to send it now.
        if (result == dfx::CollectorState::ChunkReady) {
            sendNetworkChunk(collector, "payload");
        }

        // Anotate frame and draw it
        auto annotated = annotateImage(image, face);
    }

    auto endTime = std::chrono::high_resolution_clock::now();
    auto totalTime = std::chrono::duration_cast<std::chrono::microseconds>(endTime-startTime).count();
    std::cout << "TIME: " << totalTime/1000000.0 << " seconds" << std::endl;
    std::cout << "#FRAMES: " << totalFrames << std::endl;
    std::cout << "FPS RATE: " << (totalFrames/(totalTime/1000000.0)) << std::endl;

    sendNetworkChunk(collector, "payload");
}

void sendNetworkChunk(std::shared_ptr<dfx::Collector> collector, std::string name) {

    // Obtain the chunk data from the collector after which the colllector is ready to
    // start adding new frames and the chunkData can be processed by a background thread.
    auto chunkData = collector->getChunkData();

    if ( chunkData == nullptr ) {
        std::cout << "CHUNK WAS NULL!" << std::endl;
        return;
    }

    // Ideally, process the chunkData in a background thread rather then this primary thread
    auto chunk = chunkData->getChunkPayload();

    auto chunkNumber = chunk.chunkNumber;

    // The chunk.payload should really be sent to the DFX Server to obtain the results,
    // but we write them here to a file to show how to get the payload.

    std::cout << "Saving chunk:" << chunkNumber << ", Byte Size: " << chunk.payload.size() << std::endl;

    json properties;
    properties["startFrame"]     = chunk.startFrame;
    properties["endFrame"]       = chunk.endFrame;
    properties["chunkNumber"]    = chunk.chunkNumber;
    properties["numberChunks"]   = chunk.numberChunks;
    properties["startTime_s"]    = chunk.startTime_s;
    properties["endTime_s"]      = chunk.endTime_s;
    properties["duration_s"]     = chunk.duration_s;

    std::ofstream propertiesFile;
    propertiesFile.open( "properties" + std::to_string(chunkNumber) + ".json", std::ios::out | std::ios::binary );
    propertiesFile << properties << std::endl;
    propertiesFile.close();

    std::ofstream payloadFile;
    payloadFile.open( "payload" + std::to_string(chunkNumber) + ".bin", std::ios::out | std::ios::binary );
    payloadFile.write( (char*)&(chunk.payload[0]), chunk.payload.size());
    payloadFile.close();

    std::ofstream metadataFile;
    metadataFile.open( "metadata" + std::to_string(chunkNumber) + ".bin", std::ios::out | std::ios::binary );
    metadataFile.write( (char*)&(chunk.metadata[0]), chunk.metadata.size());
    metadataFile.close();
}

std::map<std::string, dfx::Face> loadFacePoints(const std::string& jsonDataFile) {
    // read a JSON file
    std::ifstream input(jsonDataFile);
    json jsonFile;
    input >> jsonFile;

    std::map<std::string, dfx::Face> faces;
    for (auto it = jsonFile["frames"].begin(); it != jsonFile["frames"].end(); it++) {
        std::string frameNumber = it.key();
        faces[frameNumber] = decodeJsonFace(it.value());
    }
    return faces;
}

cv::Mat annotateImage(cv::Mat& image, dfx::Face& face) {
    return image.clone();
}

dfx::Face decodeJsonFace(json& jsonFace) {
    int rectX = jsonFace["rect.x"];
    int rectY = jsonFace["rect.y"];
    int rectW = jsonFace["rect.w"];
    int rectH = jsonFace["rect.h"];

    dfx::Face face;
    face.id = jsonFace["id"];
    face.faceRect = cv::Rect(rectX, rectY, rectW, rectH);
    face.poseValid = jsonFace["poseValid"];
    face.detected = jsonFace["detected"];
    for (auto it = jsonFace["points"].begin(); it != jsonFace["points"].end(); it++) {
            std::string pointName(it.key());
            auto jsonPoint = it.value();

            dfx::PosePoint posePoint;
            posePoint.point.x = jsonPoint["x"];
            posePoint.point.y = jsonPoint["y"];
            posePoint.point.z = 0;    // Safe to ignore
            posePoint.valid = jsonPoint["valid"];
            posePoint.estimated = jsonPoint["estimated"];
            posePoint.quality = jsonPoint["quality"];

            face.posePoints.emplace(pointName, posePoint);
    }
    return face;
}

