/*
 *              Copyright (c) 2016-2019, Nuralogix Corp.
 *                      All Rights reserved
 *
 *      THIS SOFTWARE IS LICENSED BY AND IS THE CONFIDENTIAL AND
 *      PROPRIETARY PROPERTY OF NURALOGIX CORP. IT IS
 *      PROTECTED UNDER THE COPYRIGHT LAWS OF THE USA, CANADA
 *      AND OTHER FOREIGN COUNTRIES. THIS SOFTWARE OR ANY
 *      PART THEREOF, SHALL NOT, WITHOUT THE PRIOR WRITTEN CONSENT
 *      OF NURALOGIX CORP, BE USED, COPIED, DISCLOSED,
 *      DECOMPILED, DISASSEMBLED, MODIFIED OR OTHERWISE TRANSFERRED
 *      EXCEPT IN ACCORDANCE WITH THE TERMS AND CONDITIONS OF A
 *      NURALOGIX CORP SOFTWARE LICENSE AGREEMENT.
 */
#include "dfxcpp/CppCollector.h"
#include "dfxcpp/CppChunkData.h"
#include "dfxcpp/CppMeasurementResult.h"
#include "dfxcpp/CppFrame.h"

#include "dfx/DFXErrors.h"

#include "dfxc/dfx_chunk_data.h"
#include "dfxc/dfx_collector.h"
#include "dfxc/dfx_measurement_result.h"
#include "dfxc/dfx_frame.h"

using dfxcpp::CppCollector;
using dfx::CollectorState;
using dfx::ThrowOnError;

CppCollector::CppCollector(std::unique_ptr<dfx_collector, void(*)(dfx_collector*)> collector) :
	dfxCollector(std::move(collector))
{

}

CppCollector::~CppCollector() {
	// Nothing to do - cleared by unique_ptr
}

std::shared_ptr<dfx::Frame> CppCollector::createFrame(const dfx::VideoFrame &videoFrame) {

	// NOTE: videoFrame has a cv::Mat which is reference counted so
	// we store a reference in CppFrame -- so we can ensure the
	// image.data is around as long as our Frame without physically
	// copying and adhering to the ABI types.

    dfx_video_frame video_frame;
	video_frame.rows = videoFrame.image.rows;
	video_frame.cols = videoFrame.image.cols;
	video_frame.pixel_type = videoFrame.image.type();
	video_frame.stride = videoFrame.image.step;
	video_frame.data = videoFrame.image.data;
	video_frame.order = static_cast<int>(videoFrame.order);
	video_frame.timestamp_ns = videoFrame.timestamp_ns;
	video_frame.number = videoFrame.number;
	video_frame.timestamp_millisec = videoFrame.timestamp_millisec;

	auto dfxFrame = dfx_frame_create(&(*dfxCollector), &video_frame, ThrowOnError{});

	// If dfx_frame_create fails it may return nullptr which occurs generally
	// when the frame state does not match what the collector anticipates.
	// ie. frame numbers/timestamps increasing, and constant pixel_type/order.
	if ( dfxFrame == nullptr ) {
		return nullptr;
	}

	return std::make_shared<dfxcpp::CppFrame>(&(*dfxCollector), videoFrame,
		std::unique_ptr<dfx_frame, void(*)(dfx_frame*)>(dfxFrame,
						[](dfx_frame* f){ dfx_frame_destroy(f); }));
}

dfx::CollectorState CppCollector::startCollection() {
	return static_cast<dfx::CollectorState>(dfx_collector_start_collection(&(*dfxCollector), ThrowOnError{}));
}

dfx::CollectorState CppCollector::getCollectorState() {
	return static_cast<dfx::CollectorState>(dfx_collector_get_state(&(*dfxCollector), ThrowOnError{}));
}

bool CppCollector::prepareMeasurement(const std::vector<uint8_t>& data) {
	bool result = dfx_collector_prepare_measurement(&(*dfxCollector),
		data.size(), data.data(), ThrowOnError{});
	return result;
}

std::string CppCollector::getMode() const {
	std::string result;
	auto setValue = [&result](const char* mode) { result = mode; };
	using T = decltype(&setValue);
	auto callback = [](void* client_data, const char* mode) {
		auto fn = static_cast<T>(client_data);
		(*fn)(mode);
	};
	dfx_collector_get_mode(&(*dfxCollector), callback, &setValue, ThrowOnError{});
	return result;
}

bool CppCollector::setTargetFPS(float targetFPS) {
	return dfx_collector_set_target_fps(&(*dfxCollector), targetFPS, ThrowOnError{});
}

bool CppCollector::setChunkDurationSeconds(float chunkDuration) {
	return dfx_collector_set_chunk_duration_seconds(&(*dfxCollector), chunkDuration, ThrowOnError{});
}

float CppCollector::getChunkDurationSeconds() {
	return dfx_collector_get_chunk_duration_seconds(&(*dfxCollector), ThrowOnError{});
}

bool CppCollector::setNumberChunks(uint32_t numberChunks) {
	return dfx_collector_set_number_chunks(&(*dfxCollector), numberChunks, ThrowOnError{});
}

uint32_t CppCollector::getNumberChunks() {
	return dfx_collector_get_number_chunks(&(*dfxCollector), ThrowOnError{});
}

bool CppCollector::enableConstraint(const std::string &constraintID) {
	return dfx_collector_enable_constraint(&(*dfxCollector), constraintID.c_str(), ThrowOnError{});
}

bool CppCollector::disableConstraint(const std::string &constraintID) {
	return dfx_collector_disable_constraint(&(*dfxCollector), constraintID.c_str(), ThrowOnError{});
}

std::vector<std::string> CppCollector::getEnabledConstraints() {
	std::vector<std::string> constraints;
    auto addResult = [&constraints](const char* name) {
        constraints.push_back(std::string(name));
    };
	using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* name) {
        auto fn = static_cast<T>(client_data);
        (*fn)(name);
    };
	dfx_collector_get_enabled_constraints(&(*dfxCollector), callback, &addResult, ThrowOnError{});
	return constraints;
}

std::vector<std::string> CppCollector::getAvailableConstraints() const {
	std::vector<std::string> constraints;
    auto addResult = [&constraints](const char* name) {
        constraints.push_back(std::string(name));
    };
	using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* name) {
        auto fn = static_cast<T>(client_data);
        (*fn)(name);
    };
	dfx_collector_get_available_constraints(&(*dfxCollector), callback, &addResult, ThrowOnError{});
	return constraints;
}

dfx::ConstraintResult CppCollector::checkConstraints(std::shared_ptr<dfx::Frame> &frame, std::vector<std::pair<std::string,dfx::ConstraintResult>> &results) {

	auto addResult = [&results](const char* constraint, uint8_t status) {
		dfx::ConstraintResult result = CppCollector::_convertToConstraintResult(status);
        results.emplace_back(std::string(constraint), result);
    };
	using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* constraint, uint8_t status) {
        auto fn = static_cast<T>(client_data);
        (*fn)(constraint, status);
    };

    CppFrame* cppFrame = static_cast<CppFrame*>(&(*frame));
	auto result = dfx_collector_check_constraints(&(*dfxCollector),
			cppFrame->_getDFXFrame(), callback, &addResult, ThrowOnError{});

	return CppCollector::_convertToConstraintResult(result);
}

dfx::ConstraintResult CppCollector::_convertToConstraintResult(uint8_t result) {
	if ( result > 2 ) { result = 2; }
	return static_cast<dfx::ConstraintResult>(result);
}

std::string CppCollector::getConstraintErrorMessage(const std::string &violationID) {
	std::string errorMessage;
	auto setValue = [&errorMessage](const char* message) { errorMessage = message; };
	using T = decltype(&setValue);
	auto callback = [](void* client_data, const char* prop) {
		auto fn = static_cast<T>(client_data);
		(*fn)(prop);
	};
	dfx_collector_get_constraint_error_message(&(*dfxCollector), violationID.c_str(), callback, &setValue, ThrowOnError{});
	return errorMessage;
}

bool CppCollector::setConstraintsConfig(const std::string& key, const std::string &propertyValue) {
	return dfx_collector_set_constraints_config(&(*dfxCollector), key.c_str(), propertyValue.c_str(), ThrowOnError{});
}

std::string CppCollector::getConstraintsConfig(const std::string& key) {
	std::string constraintConfig;
	auto setValue = [&constraintConfig](const char* config) { constraintConfig = config; };
	using T = decltype(&setValue);
	auto callback = [](void* client_data, const char* prop) {
		auto fn = static_cast<T>(client_data);
		(*fn)(prop);
	};
	dfx_collector_get_constraints_config(&(*dfxCollector), key.c_str(), callback, &setValue, ThrowOnError{});
	return constraintConfig;
}

std::vector<std::string> CppCollector::getRequiredPosePointIDs() const {
	std::vector<std::string> posePointIDs;
    auto addResult = [&posePointIDs](const char* name) {
        posePointIDs.push_back(std::string(name));
    };
	using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* name) {
        auto fn = static_cast<T>(client_data);
        (*fn)(name);
    };
	dfx_collector_get_required_pose_point_ids(&(*dfxCollector), callback, &addResult, ThrowOnError{});
	return posePointIDs;
}

dfx::CollectorState CppCollector::defineRegions(std::shared_ptr<dfx::Frame> &frame) {
	CppFrame* cppFrame = static_cast<CppFrame*>(&(*frame));
	return static_cast<dfx::CollectorState>(dfx_collector_define_regions(&(*dfxCollector), cppFrame->_getDFXFrame(), ThrowOnError{}));
}

dfx::CollectorState CppCollector::extractChannels(std::shared_ptr<dfx::Frame> &frame) {
	CppFrame* cppFrame = static_cast<CppFrame*>(&(*frame));
	return static_cast<dfx::CollectorState>(dfx_collector_extract_channels(&(*dfxCollector), cppFrame->_getDFXFrame(), ThrowOnError{}));
}

bool CppCollector::isChunkReady() const {
	return dfx_collector_is_chunk_ready(&(*dfxCollector), ThrowOnError{});
}

int32_t CppCollector::numberFramesNeeded() const {
	return dfx_collector_number_frames_needed(&(*dfxCollector), ThrowOnError{});
}

std::shared_ptr<dfx::ChunkData> CppCollector::getChunkData() {
	auto chunkData = dfx_collector_get_chunk_data(&(*dfxCollector), ThrowOnError{});

	// Don't wrap a null pointer
	if ( ! chunkData ) {
		return nullptr;
	}

	return std::make_shared<dfxcpp::CppChunkData>(
        std::unique_ptr<dfx_chunk_data, void(*)(dfx_chunk_data*)>(
            chunkData, 
            [](dfx_chunk_data* c) { dfx_chunk_data_destroy(c); }));
}

std::shared_ptr<dfx::MeasurementResult> CppCollector::decodeMeasurementResult(const std::vector<uint8_t>& measurementResponse) {
	uint32_t size = measurementResponse.size();
	auto measurementResult = dfx_collector_decode_measurement_result(
		&(*dfxCollector), size, &measurementResponse.front(), ThrowOnError{});
    return std::make_shared<dfxcpp::CppMeasurementResult>(
        std::unique_ptr<dfx_measurement_result, void(*)(dfx_measurement_result*)>(
            measurementResult, 
            [](dfx_measurement_result* r) { dfx_measurement_result_destroy(r); }));
}

bool CppCollector::setProperty(const std::string &key, const std::string &value) {
	return dfx_collector_set_property(&(*dfxCollector), key.c_str(), value.c_str(), ThrowOnError{});
}

std::string CppCollector::getProperty(const std::string& key) {
	std::string property;
	auto setValue = [&property](const char* prop) { property = prop; };
	using T = decltype(&setValue);
	auto callback = [](void* client_data, const char* prop) {
		auto fn = static_cast<T>(client_data);
		(*fn)(prop);
	};
	dfx_collector_get_property(&(*dfxCollector), key.c_str(), callback, &setValue, ThrowOnError{});
	return property;
}

void CppCollector::removeProperty(const std::string& key) {
	dfx_collector_remove_property(&(*dfxCollector), key.c_str(), ThrowOnError{});
}

std::map<std::string, std::string> CppCollector::getProperties() {
	std::map<std::string, std::string> props;

    auto addResult = [&props](const char* key, const char* value) {
        props.emplace(std::string(key), std::string(value));
    };
	using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* key, const char* value) {
        auto fn = static_cast<T>(client_data);
        (*fn)(key, value);
    };
    dfx_collector_get_properties(&(*dfxCollector), callback, &addResult, ThrowOnError{});
	return props;
}

std::string CppCollector::getLastErrorMessage() {
	std::string lastErrorMessage;
	auto setValue = [&lastErrorMessage](const char* message) { lastErrorMessage = message; };
	using T = decltype(&setValue);
	auto callback = [](void* client_data, const char* message) {
		auto fn = static_cast<T>(client_data);
		(*fn)(message);
	};
	dfx_collector_get_last_error_message(&(*dfxCollector), callback, &setValue, ThrowOnError{});
	return lastErrorMessage;
}

void CppCollector::cancelCollection() {
	dfx_collector_cancel_collection(&(*dfxCollector), ThrowOnError{});
}

void CppCollector::stopCollection() {
	cancelCollection();
}

void CppCollector::forceComplete() {
	dfx_collector_force_complete(&(*dfxCollector), ThrowOnError{});
}

void CppCollector::resetCollection() {
	dfx_collector_reset_collection(&(*dfxCollector), ThrowOnError{});
}
