/*
 *              Copyright (c) 2016-2019, Nuralogix Corp.
 *                      All Rights reserved
 *
 *      THIS SOFTWARE IS LICENSED BY AND IS THE CONFIDENTIAL AND
 *      PROPRIETARY PROPERTY OF NURALOGIX CORP. IT IS
 *      PROTECTED UNDER THE COPYRIGHT LAWS OF THE USA, CANADA
 *      AND OTHER FOREIGN COUNTRIES. THIS SOFTWARE OR ANY
 *      PART THEREOF, SHALL NOT, WITHOUT THE PRIOR WRITTEN CONSENT
 *      OF NURALOGIX CORP, BE USED, COPIED, DISCLOSED,
 *      DECOMPILED, DISASSEMBLED, MODIFIED OR OTHERWISE TRANSFERRED
 *      EXCEPT IN ACCORDANCE WITH THE TERMS AND CONDITIONS OF A
 *      NURALOGIX CORP SOFTWARE LICENSE AGREEMENT.
 */
#include "dfxcpp/CppFrame.h"
#include "dfxcpp/CppCollector.h"

#include "dfx/DFXErrors.h"

#include "dfxc/dfx_face.h"
#include "dfxc/dfx_frame.h"

using dfxcpp::CppFrame;
using dfx::CollectorState;
using dfx::ThrowOnError;
using dfx::VideoFrame;

CppFrame::~CppFrame() = default;

CppFrame::CppFrame(dfx_collector* dfxCollector, VideoFrame videoFrame, std::unique_ptr<dfx_frame, void(*)(dfx_frame*)> frame)
    : collector(dfxCollector),
      videoFrame(videoFrame),
      dfxFrame(std::move(frame)) {
}

dfx_frame* CppFrame::_getDFXFrame() {
	return &(*dfxFrame);
}

VideoFrame CppFrame::getVideoFrame() const  {
    return videoFrame;
}

bool CppFrame::addFace(std::shared_ptr<dfx::Face> face) {
    bool result = false;

	auto dfxFace = dfx_face_create(face->id.c_str(), ThrowOnError{});
	assert(dfxFace != nullptr);
	if ( dfxFace != nullptr ) {
		auto& rect = face->faceRect;
		dfx_face_set_face_rect(dfxFace, rect.x, rect.y, rect.width, rect.height, ThrowOnError{});
		dfx_face_set_pose_valid(dfxFace, face->poseValid, ThrowOnError{});
		dfx_face_set_detected(dfxFace, face->detected, ThrowOnError{});

		// Add Pose Points
		for (auto& posePoint : face->posePoints ) {
			auto& point_name = posePoint.first;
			auto& point = posePoint.second;
			dfx_face_add_pose_point(dfxFace, point_name.c_str(),
				point.point.x, point.point.y, point.point.z,
				point.valid,   point.estimated, point.quality,
				ThrowOnError{});
		}

		// Add Attributes
		for (auto& attribute : face->attributes ) {
			auto& name = attribute.first;
			auto& value = attribute.second;
			dfx_face_add_attribute(dfxFace, name.c_str(), value, ThrowOnError{});
		}

		result = dfx_frame_add_face(&(*dfxFrame), dfxFace, ThrowOnError{});

		dfx_face_destroy(dfxFace);
	}

    return result;
}

std::vector<float> CppFrame::getQualityMetrics(float iso, float exposure) {
    std::vector<float> metrics;
    metrics.resize(64);

    int number_metrics = 0;
    auto result = dfx_frame_get_quality_metrics(&(*dfxFrame),
                                          iso, exposure, metrics.capacity(),
                                          &metrics[0], &number_metrics);
    if ( result != 0 ) {
        metrics.resize(number_metrics);
        return metrics;
    }

    return std::vector<float>();
}

int CppFrame::getStarRating(std::string& feedback) {
    int rating = 0;
    char response[256];
    dfx_frame_get_star_rating(&(*dfxFrame), sizeof(response), response, &rating);
    feedback = std::string(response);
    return rating;
}

bool CppFrame::addMarker(const std::string &label) {
    return dfx_frame_add_marker(&(*dfxFrame), label.c_str(), ThrowOnError{});
}

std::vector<std::string> CppFrame::getMarkers() const {
    std::vector<std::string> markers;

    auto addResult = [&markers](const char* marker) {
        markers.push_back(std::string(marker));
    };
    using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* marker) {
        auto fn = static_cast<T>(client_data);
        (*fn)(marker);
    };
    dfx_frame_get_markers(&(*dfxFrame), callback, &addResult, ThrowOnError{});

    return markers;
}

std::vector<std::string> CppFrame::getDesiredFaceAttributes() const {
    std::vector<std::string> desiredAttributes;

    auto addResult = [&desiredAttributes](const char* name) {
        desiredAttributes.push_back(std::string(name));
    };
    using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* name) {
        auto fn = static_cast<T>(client_data);
        (*fn)(name);
    };
    dfx_frame_get_desired_face_attributes(&(*dfxFrame), callback, &addResult, ThrowOnError{});

    return desiredAttributes;
}

std::vector<std::string> CppFrame::getDesiredDeviceAttributes() const {
    std::vector<std::string> desiredAttributes;

    auto addResult = [&desiredAttributes](const char* name) {
        desiredAttributes.push_back(std::string(name));
    };
    using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* name) {
        auto fn = static_cast<T>(client_data);
        (*fn)(name);
    };
    dfx_frame_get_desired_device_attributes(&(*dfxFrame), callback, &addResult, ThrowOnError{});

    return desiredAttributes;
}

double CppFrame::getDeviceAttribute(const std::string& key) const {
    return dfx_frame_get_device_attribute(&(*dfxFrame), key.c_str(), ThrowOnError{});
}

void CppFrame::setDeviceAttribute(const std::string& key, double value) {
    dfx_frame_set_device_attribute(&(*dfxFrame), key.c_str(), value, ThrowOnError{});
}

std::vector<std::string> CppFrame::getFaceIdentifiers() const {
    std::vector<std::string> faceIds;

    auto addResult = [&faceIds](const char* name) {
        faceIds.push_back(std::string(name));
    };
    using T = decltype(&addResult);
    auto callback = [](void* client_data, const char* name) {
        auto fn = static_cast<T>(client_data);
        (*fn)(name);
    };
	dfx_frame_get_face_identifiers(&(*dfxFrame), callback, &addResult, ThrowOnError{});

    return faceIds;
};

std::vector<std::string> CppFrame::getRegionNames(const std::string &faceID) const {
    std::vector<std::string> roiNames;

    auto addResult = [&roiNames](const char* name) {
        roiNames.push_back(std::string(name));
    };
    using T  = decltype(&addResult);
    auto callback = [](void* client_data, const char* name) {
        auto fn = static_cast<T>(client_data);
        (*fn)(name);
    };

    dfx_frame_get_region_names(&(*dfxFrame), faceID.c_str(), callback, &addResult, ThrowOnError{});

    return roiNames;
}

std::vector<cv::Point> CppFrame::getRegionPolygon(const std::string &faceID, const std::string &regionName) const {
    std::vector<cv::Point> roiPolygon;

    auto addResult = [&roiPolygon](float x, float y) {
        roiPolygon.push_back(cv::Point(x, y));
    };
    using T = decltype(&addResult);
    auto callback = [](void* client_data, float x, float y) {
        auto fn = static_cast<T>(client_data);
        (*fn)(x, y);
    };

 	dfx_frame_get_region_polygon(&(*dfxFrame), faceID.c_str(), regionName.c_str(), callback, &addResult, ThrowOnError{});

    return roiPolygon;
}

int32_t CppFrame::getRegionIntProperty(const std::string& faceID, const std::string& regionName, const std::string& property) const {
    return dfx_frame_get_region_int_property(&(*dfxFrame), faceID.c_str(), regionName.c_str(), property.c_str(), ThrowOnError{});
}

cv::Mat CppFrame::getRegionHistogram(const std::string &faceID, const std::string &regionName) const {
    cv::Mat mat;

    auto addResult = [&mat](uint16_t rows, uint16_t cols, uint8_t pixel_type,
						size_t stride, const uint8_t* data,	uint8_t order) {
        mat = cv::Mat( rows, cols, pixel_type, (void*)data, stride );
    };
    using T = decltype(&addResult);
    auto callback = [](void* client_data, uint16_t rows, uint16_t cols,	uint8_t pixel_type,
						size_t stride, const uint8_t* data,	uint8_t order) {
        auto fn = static_cast<T>(client_data);
        (*fn)(rows, cols, pixel_type, stride, data, order);
    };

	dfx_frame_get_region_histogram(&(*dfxFrame), faceID.c_str(), regionName.c_str(), callback, &addResult, ThrowOnError{});

    return mat;
}

